<?php
if ( ! defined( 'ABSPATH' ) ) {
	die();
}// Exit if accessed directly

/**
 * Bitcore Insight Block Explorer API Class (supports Bitcore Insight version 8)
 *
 * @category   CryptoWoo
 * @package    OrderProcessing
 * @subpackage BlockExplorerAPI
 * @author     CryptoWoo AS
 */
abstract class CW_Block_Explorer_API_Bitcore_Insight extends CW_Block_Explorer_Base {

	/**
	 *
	 * Get the formatting of currency pair for exchange API
	 *
	 * @return string
	 */
	protected function get_txs_endpoint_format() : string {
		return '%1$s/mainnet/address/%2$s/txs';
	}

	/**
	 *
	 * Get the block explorer api block height endpoint format
	 *
	 * @return string
	 */
	protected function get_block_height_endpoint_format() : string {
		return '%1$s/mainnet/block/tip';
	}

	/**
	 *
	 * Get the block explorer block hash api endpoint format
	 *
	 * @return string
	 */
	protected function get_block_hash_endpoint_format() : string {
		return '%1$s/mainnet/block/%2$d';
	}

	/**
	 *
	 * Get the block explorer txs key name
	 *
	 * @return string
	 */
	protected function get_txs_key_name() : string {
		return '';
	}

	/**
	 *
	 * Get the block explorer block height key name
	 *
	 * @return string
	 */
	protected function get_block_height_key_name() : string {
		return 'height';
	}

	/**
	 *
	 * Get the block explorer block hash key name
	 *
	 * @return string
	 */
	protected function get_block_hash_key_name() : string {
		return 'hash';
	}

	/**
	 *
	 * Get the block explorer txs txid key name
	 *
	 * @return string
	 */
	protected function get_tx_txid_key_name() : string {
		return 'mintTxid';
	}

	/**
	 *
	 * Get the block explorer txs confirms key name
	 *
	 * @return array
	 */
	protected function get_tx_confirms_key_name() : string {
		return 'confirmations';
	}

	/**
	 *
	 * Get the block explorer txs amount key name
	 *
	 * @return string
	 */
	protected function get_tx_amount_key_name() : string {
		return 'value';
	}

	/**
	 *
	 * Get the block explorer txs locktime key name
	 *
	 * @return string
	 */
	protected function get_tx_locktime_key_name() : string {
		return '';
	}

	/**
	 *
	 * Get the block explorer txs timestamp key name
	 *
	 * @return string
	 */
	protected function get_tx_timestamp_key_name() : string {
		return 'time';
	}

	/**
	 *
	 * Get the block explorer tx block height key name
	 *
	 * @return string
	 */
	protected function get_tx_block_height_key_name() : string {
		return 'mintHeight';
	}

	/**
	 *
	 * Get the block explorer tx double spend key name
	 *
	 * @return string
	 */
	protected function get_tx_double_spend_key_name() : string {
		return '';
	}

	/**
	 *
	 * Get the block explorer tx inputs key name
	 *
	 * @return string
	 */
	protected function get_tx_inputs_key_name() : string {
		return '';
	}

	/**
	 *
	 * Get the block explorer tx outputs key name
	 *
	 * @return string
	 */
	protected function get_tx_outputs_key_name() : string {
		return '';
	}

	/**
	 *
	 * Get the block explorer tx input sequence key name
	 *
	 * @return string
	 */
	protected function get_tx_input_sequence_key_name() : string {
		return 'sequenceNumber';
	}

	/**
	 *
	 * Get the block explorer txs address key name
	 *
	 * @return string
	 */
	protected function get_tx_address_key_name() : string {
		return 'address';
	}

	/**
	 *
	 * Is the block explorer currency uppercase or lowercase in the api url?
	 * Default is lower case.
	 *
	 * @return bool
	 */
	protected function search_currency_is_uppercase(): bool {
		return true;
	}

	/**
	 *
	 * Is the block explorer crypto amount in satoshi (1e8)?
	 * Default is true (amount is satoshi ).
	 *
	 * @return bool
	 */
	protected function amount_from_api_is_satoshi(): bool {
		return true;
	}

	/**
	 *
	 * Get the block explorer max txs allowed in api call
	 *
	 * @return int
	 */
	public function get_api_max_allowed_addresses() : int {
		return 1;
	}


	/**
	 *
	 * Format the data from block explorer result to default data format
	 *
	 * @param stdClass|array $block_explorer_data Json decoded result from block explorer api call.
	 *
	 * @return stdClass|array
	 */
	protected function format_result_from_block_explorer( $block_explorer_data ) {
		if ( ! is_array( $block_explorer_data ) ) {
			return $block_explorer_data;
		}

		foreach ( $block_explorer_data as $txs_data ) {
			if ( isset( $txs_data->{$this->get_tx_confirms_key_name()} ) && - 1 === $txs_data->{$this->get_tx_confirms_key_name()} ) {
				$txs_data->{$this->get_tx_confirms_key_name()} = null;
			}
			// TODO: Find the timestamp eg. by getting it from the time of the block height of tx?
			$txs_data->{$this->get_tx_timestamp_key_name()} = false;
		}

		return $block_explorer_data;
	}

	/**
	 *
	 * Get the address or addresses string for use in blockchain url parameter or url
	 *
	 * @return string
	 * TODO: Remove this when cash address is used by default for Bitcoin Cash.
	 */
	protected function get_addresses_string(): string {
		if ( 'BCH' === $this->get_search_currency() && is_array( $this->get_addresses_array() ) ) {
			$cash_addresses = array();
			foreach ( $this->get_addresses_array() as & $address ) {
				if ( file_exists( CWOO_PLUGIN_DIR . 'includes/class-cashaddress.php' ) ) {
					include_once CWOO_PLUGIN_DIR . 'includes/class-cashaddress.php';
					$cash_addresses[] = str_replace( 'bitcoincash:', '', \CashAddress\CashAddress::old2new( $address ) );
				}

				return implode( ',', $cash_addresses );
			}
		}

		return parent::get_addresses_string();
	}

	/**
	 * Format data result nicely for array dataset.
	 *
	 * @param string $key_name Key name in resultset.
	 * @param mixed  $data Data in resultset.
	 *
	 * @return array
	 *
	 * TODO: Remove this when cash address is used by default for Bitcoin Cash.
	 */
	protected function format_data_result_array( $key_name, $data ) : array {
		$result = parent::format_data_result_array( $key_name, $data );

		if ( 'BCH' !== $this->get_search_currency() ) {
			return $result;
		}

		if ( isset( $result['txs'] ) && is_array( $result['txs'] ) ) {
			foreach ( $result['txs'] as $address => $tx ) {
				if ( 0 === strpos( $address, 'q' ) ) {
					$cash_address = \CashAddress\CashAddress::new2old( $address, false );
					if ( $cash_address === $this->get_current_address() ) {
						$result['txs'][ $cash_address ] = $tx;
						unset( $result['txs'][ $address ] );
					}
				}
			}
		}

		return $result;
	}
}
